@@ -1,6 +1,5 @@
 module Test.MockCat
-  ( 
-    module Test.MockCat.Mock,
+  ( module Test.MockCat.Mock,
     module Test.MockCat.Cons,
     module Test.MockCat.Param,
     module Test.MockCat.MockT,
