@@ -2,10 +2,13 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE FunctionalDependencies #-}
 {-# LANGUAGE NamedFieldPuns #-}
+{-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE TypeOperators #-}
 {-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
 {-# HLINT ignore "Use null" #-}
 {-# OPTIONS_GHC -Wno-name-shadowing #-}
+{-# LANGUAGE UndecidableInstances #-}
+{-# LANGUAGE InstanceSigs #-}
 
 {- | This module provides the following functions.
 
@@ -45,6 +48,7 @@ where
 
 import Control.Monad (guard, when, ap)
 import Data.Function ((&))
+import Data.Char (isLower)
 import Data.IORef (IORef, newIORef, readIORef, atomicModifyIORef')
 import Data.List (elemIndex, intercalate)
 import Data.Maybe
@@ -52,14 +56,13 @@ import Data.Text (pack, replace, unpack)
 import GHC.IO (unsafePerformIO)
 import Test.MockCat.Cons
 import Test.MockCat.Param
-import Test.MockCat.ParamDivider
 import Test.MockCat.AssociationList (AssociationList, lookup, update, insert, empty, member)
 import Prelude hiding (lookup)
 import GHC.Stack (HasCallStack)
 import Control.Monad.Trans
 import Control.Monad.State
 
-data Mock fun params = Mock (Maybe MockName) fun (Verifier params)
+data Mock fn params = Mock (Maybe MockName) fn (Verifier params)
 
 type MockName = String
 
@@ -87,10 +90,9 @@ From this mock, you can generate stub functions and verify the functions.
 
 -}
 createMock ::
-  MockBuilder params fun verifyParams =>
-  MonadIO m =>
+  (MonadIO m, MockBuilder params fn verifyParams) =>
   params ->
-  m (Mock fun verifyParams)
+  m (Mock fn verifyParams)
 createMock params = liftIO $ build Nothing params
 
 {- | Create a constant mock.
@@ -121,19 +123,18 @@ createConstantMock a = liftIO $ build Nothing $ param a
   @
 -}
 createNamedMock ::
-  MockBuilder params fun verifyParams =>
-  MonadIO m =>
+  (MonadIO m, MockBuilder params fn verifyParams) =>
   MockName ->
   params ->
-  m (Mock fun verifyParams)
+  m (Mock fn verifyParams)
 createNamedMock name params = liftIO $ build (Just name) params
 
 -- | Create a named constant mock.
-createNamedConstantMock :: MonadIO m => MockName -> fun -> m (Mock fun ())
+createNamedConstantMock :: MonadIO m => MockName -> fn -> m (Mock fn ())
 createNamedConstantMock name a = liftIO $ build (Just name) (param a)
 
 -- | Extract the stub function from the mock.
-stubFn :: Mock fun v -> fun
+stubFn :: Mock fn v -> fn
 stubFn (Mock _ f _) = f
 
 {- | Create a stub function.
@@ -147,142 +148,52 @@ stubFn (Mock _ f _) = f
   @
 -}
 createStubFn ::
-  MockBuilder params fun verifyParams =>
-  MonadIO m =>
+  (MonadIO m, MockBuilder params fn verifyParams) =>
   params ->
-  m fun
+  m fn
 createStubFn params = stubFn <$> createMock params
 
 -- | Create a named stub function.
 createNamedStubFn ::
-  MockBuilder params fun verifyParams =>
-  MonadIO m =>
+  (MonadIO m, MockBuilder params fn verifyParams) =>
   String ->
   params ->
-  m fun
+  m fn
 createNamedStubFn name params = stubFn <$> createNamedMock name params
 
+-- | Class for building a curried function.
+-- The purpose of this class is to automatically generate and provide
+-- an implementation for the corresponding curried function type (such as `a -> b -> ... -> IO r`)
+-- when given the argument list type of the mock (`Param a :> Param b :> ...`).
+-- @args@ is the argument list type of the mock.
+-- @r@ is the return type of the function.
+-- @fn@ is the curried function type.
+class BuildCurried args r fn | args r -> fn where
+  -- | Build a curried function.
+  -- Accept a function that combines all arguments and convert it into a curried function.
+  buildCurried :: (args -> IO r) -> fn
+
+-- | Base case: The last parameter.
+-- Converts a single-argument function (Param a -> IO r) into a final
+-- curried function (a -> r) by performing the IO action.
+instance BuildCurried (Param a) r (a -> r) where
+  buildCurried :: (Param a -> IO r) -> a -> r
+  buildCurried a2r a = perform $ a2r (param a)
+
+-- | Recursive case: Consumes the head parameter and generates the next curried function.
+-- Generates a function of type (a -> fn) that immediately calls the next
+-- 'BuildCurried' instance with the remaining arguments.
+instance BuildCurried rest r fn
+      => BuildCurried (Param a :> rest) r (a -> fn) where
+  buildCurried :: ((Param a :> rest) -> IO r) -> a -> fn
+  buildCurried args2r a = buildCurried (\rest -> args2r (p a :> rest))
+
 -- | Class for creating a mock corresponding to the parameter.
-class MockBuilder params fun verifyParams | params -> fun, params -> verifyParams where
+class MockBuilder params fn verifyParams | params -> fn, params -> verifyParams where
   -- build a mock
-  build :: MonadIO m => Maybe MockName -> params -> m (Mock fun verifyParams)
-
--- instances
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g, Show h, Eq h, Show i, Eq i) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param i :> Param r)
-    (a -> b -> c -> d -> e -> f -> g -> h -> i -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param i)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock
-      name
-      s
-      ( \a2 b2 c2 d2 e2 f2 g2 h2 i2 ->
-          unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2 :> p h2 :> p i2) s
-      )
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g, Show h, Eq h) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param r)
-    (a -> b -> c -> d -> e -> f -> g -> h -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock
-      name
-      s
-      ( \a2 b2 c2 d2 e2 f2 g2 h2 ->
-          unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2 :> p h2) s
-      )
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param r)
-    (a -> b -> c -> d -> e -> f -> g -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock
-      name
-      s
-      ( \a2 b2 c2 d2 e2 f2 g2 ->
-          unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2) s
-      )
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param r)
-    (a -> b -> c -> d -> e -> f -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 f2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param r)
-    (a -> b -> c -> d -> e -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d) =>
-  MockBuilder
-    (Param a :> Param b :> Param c :> Param d :> Param r)
-    (a -> b -> c -> d -> r)
-    (Param a :> Param b :> Param c :> Param d)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2 :> p d2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c) =>
-  MockBuilder (Param a :> Param b :> Param c :> Param r) (a -> b -> c -> r) (Param a :> Param b :> Param c)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2 :> p c2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b) =>
-  MockBuilder (Param a :> Param b :> Param r) (a -> b -> r) (Param a :> Param b)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2 :> p b2) s)
-
-instance
-  (Show a, Eq a) =>
-  MockBuilder (Param a :> Param r) (a -> r) (Param a)
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 -> unsafePerformIO $ extractReturnValueWithValidate name params (p a2) s)
-
-instance
-  MockBuilder (Param r) r ()
-  where
-  build name params = do
-    s <- liftIO $ newIORef appliedRecord
-    let v = value params
-    makeMock name s $ unsafePerformIO (do
-      liftIO $ appendAppliedParams s ()
-      pure v)
+  build :: MonadIO m => Maybe MockName -> params -> m (Mock fn verifyParams)
 
+-- | Instance for building a mock for a constant function.
 instance
   MockBuilder (IO r) (IO r) ()
   where
@@ -292,111 +203,19 @@ instance
       liftIO $ appendAppliedParams s ()
       a)
 
+-- | Instance for building a mock for a function with a single parameter.
 instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g, Show h, Eq h, Show i, Eq i) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param i :> Param r) ())
-    (a -> b -> c -> d -> e -> f -> g -> h -> i -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param i)
+  MockBuilder (Param r) r ()
   where
-  build name cases = do
-    let params = runCase cases
+  build name params = do
     s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 f2 g2 h2 i2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2 :> p h2 :> p i2) s)
+    let v = value params
+    makeMock name s $ perform (do
+      liftIO $ appendAppliedParams s ()
+      pure v)
 
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g, Show h, Eq h) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h :> Param r) ())
-    (a -> b -> c -> d -> e -> f -> g -> h -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param h)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 f2 g2 h2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2 :> p h2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f, Show g, Eq g) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g :> Param r) ())
-    (a -> b -> c -> d -> e -> f -> g -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param g)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 f2 g2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2 :> p g2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e, Show f, Eq f) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param e :> Param f :> Param r) ())
-    (a -> b -> c -> d -> e -> f -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e :> Param f)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 f2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2 :> p f2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d, Show e, Eq e) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param e :> Param r) ())
-    (a -> b -> c -> d -> e -> r)
-    (Param a :> Param b :> Param c :> Param d :> Param e)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 e2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2 :> p e2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c, Show d, Eq d) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param d :> Param r) ())
-    (a -> b -> c -> d -> r)
-    (Param a :> Param b :> Param c :> Param d)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 d2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2 :> p d2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b, Show c, Eq c) =>
-  MockBuilder
-    (Cases (Param a :> Param b :> Param c :> Param r) ())
-    (a -> b -> c -> r)
-    (Param a :> Param b :> Param c)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 c2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2 :> p c2) s)
-
-instance
-  (Show a, Eq a, Show b, Eq b) =>
-  MockBuilder (Cases (Param a :> Param b :> Param r) ()) (a -> b -> r) (Param a :> Param b)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 b2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2 :> p b2) s)
-
-instance
-  (Show a, Eq a) =>
-  MockBuilder (Cases (Param a :> Param r) ()) (a -> r) (Param a)
-  where
-  build name cases = do
-    let params = runCase cases
-    s <- liftIO $ newIORef appliedRecord
-    makeMock name s (\a2 -> unsafePerformIO $ findReturnValueWithStore name params (p a2) s)
-
-instance
-  MockBuilder (Cases (IO a) ()) (IO a) ()
-  where
+-- | Instance for building a mock for a function with multiple parameters.
+instance MockBuilder (Cases (IO a) ()) (IO a) () where
   build name cases = do
     let params = runCase cases
     s <- liftIO $ newIORef appliedRecord
@@ -407,31 +226,70 @@ instance
       appendAppliedParams s ()
       incrementAppliedParamCount s ()
       fromJust r)
--- ------
+
+-- | Overlapping instance for building a mock for a function with multiple parameters.
+-- This instance is used when the parameter type is a 'Cases' type.
+instance {-# OVERLAPPABLE #-}
+  ( ProjectionArgs params
+  , ProjectionReturn params
+  , ArgsOf params ~ args
+  , ReturnOf params ~ Param r
+  , BuildCurried args r fn
+  , Eq args
+  , Show args
+  ) => MockBuilder (Cases params ()) fn args where
+  build name cases = do
+    let paramsList = runCase cases
+    s <- liftIO $ newIORef appliedRecord
+    makeMock name s (buildCurried (\inputParams -> findReturnValueWithStore name paramsList inputParams s))
+
+-- | Overlapping instance for building a mock for a function with multiple parameters.
+-- This instance is used when the parameter type is a 'Param a :> rest' type.
+instance {-# OVERLAPPABLE #-}
+  ( p ~ (Param a :> rest)
+  , ProjectionArgs p
+  , ProjectionReturn p
+  , ArgsOf p ~ args
+  , ReturnOf p ~ Param r
+  , BuildCurried args r fn
+  , Eq args
+  , Show args
+  ) => MockBuilder (Param a :> rest) fn args where
+  build name params = do
+    s <- liftIO $ newIORef appliedRecord
+    makeMock name s (buildCurried (\inputParams -> extractReturnValueWithValidate name params inputParams s))
 
 p :: a -> Param a
 p = param
 
-makeMock :: MonadIO m => Maybe MockName -> IORef (AppliedRecord params) -> fun -> m (Mock fun params)
+makeMock :: MonadIO m => Maybe MockName -> IORef (AppliedRecord params) -> fn -> m (Mock fn params)
 makeMock name l fn = pure $ Mock name fn (Verifier l)
 
 extractReturnValueWithValidate ::
-  ParamDivider params args (Param r) =>
-  Eq args =>
-  Show args =>
+  ( ProjectionArgs params
+  , ProjectionReturn params
+  , ArgsOf params ~ args
+  , ReturnOf params ~ Param r
+  , Eq args
+  , Show args
+  ) =>
   Maybe MockName ->
   params ->
   args ->
   IORef (AppliedRecord args) ->
   IO r
 extractReturnValueWithValidate name params inputParams s = do
-  validateWithStoreParams name s (args params) inputParams
+  validateWithStoreParams name s (projArgs params) inputParams
   pure $ returnValue params
 
 findReturnValueWithStore ::
-  Eq args =>
-  Show args =>
-  ParamDivider params args (Param r) =>
+  ( ProjectionArgs params
+  , ProjectionReturn params
+  , ArgsOf params ~ args
+  , ReturnOf params ~ Param r
+  , Eq args
+  , Show args
+  ) =>
   Maybe MockName ->
   AppliedParamsList params ->
   args ->
@@ -439,7 +297,7 @@ findReturnValueWithStore ::
   IO r
 findReturnValueWithStore name paramsList inputParams ref = do
   appendAppliedParams ref inputParams
-  let expectedArgs = args <$> paramsList
+  let expectedArgs = projArgs <$>paramsList
   r <- findReturnValue paramsList inputParams ref
   maybe
     (errorWithoutStackTrace $ messageForMultiMock name expectedArgs inputParams)
@@ -447,14 +305,18 @@ findReturnValueWithStore name paramsList inputParams ref = do
     r
 
 findReturnValue ::
-  Eq args =>
-  ParamDivider params args (Param r) =>
+  ( ProjectionArgs params
+  , ProjectionReturn params
+  , ArgsOf params ~ args
+  , ReturnOf params ~ Param r
+  , Eq args
+  ) =>
   AppliedParamsList params ->
   args ->
   IORef (AppliedRecord args) ->
   IO (Maybe r)
 findReturnValue paramsList inputParams ref = do
-  let matchedParams = filter (\params -> args params == inputParams) paramsList
+  let matchedParams = filter (\params -> projArgs params == inputParams) paramsList
   case matchedParams of
     [] -> pure Nothing
     _ -> do
@@ -474,25 +336,50 @@ validateParams name expected actual =
     then pure ()
     else errorWithoutStackTrace $ message name expected actual
 
+-- Helper: quote a token if it looks like an unquoted alpha token
+quoteToken :: String -> String
+quoteToken s
+  | null s = s
+  | head s == '"' = s
+  | head s == '(' = s
+  | head s == '[' = s
+  | not (null s) && isLower (head s) = '"' : s ++ "\""
+  | otherwise = s
+
+-- Quote a show-produced string when appropriate for error messages.
+showForMessage :: String -> String
+showForMessage s =
+  -- if it's a parenthesised compound, keep as-is; otherwise quote alpha-only tokens
+  let trimmed = s
+   in if not (null trimmed) && head trimmed == '(' && last trimmed == ')'
+        then trimmed
+        else quoteToken trimmed
+
 message :: Show a => Maybe MockName -> a -> a -> String
 message name expected actual =
   intercalate
     "\n"
     [ "function" <> mockNameLabel name <> " was not applied to the expected arguments.",
-      "  expected: " <> show expected,
-      "   but got: " <> show actual
+      "  expected: " <> showForMessage (show expected),
+      "   but got: " <> showForMessage (show actual)
     ]
 
 messageForMultiMock :: Show a => Maybe MockName -> [a] -> a -> String
 messageForMultiMock name expecteds actual =
-  intercalate
-    "\n"
-    [ "function" <> mockNameLabel name <> " was not applied to the expected arguments.",
-      "  expected one of the following:",
-      intercalate "\n" $ ("    " <>) . show <$> expecteds,
-      "  but got:",
-      ("    " <>) . show $ actual
-    ]
+  let fmtExpected e =
+        let s = show e
+            -- if it's parenthesised compound, strip outer parens then quote inner alpha tokens
+            inner = if not (null s) && head s == '(' && last s == ')' then init (tail s) else s
+            tokens = map (trim . quoteToken . trim) (splitByComma inner)
+         in intercalate "," tokens
+   in intercalate
+        "\n"
+        [ "function" <> mockNameLabel name <> " was not applied to the expected arguments.",
+          "  expected one of the following:",
+          intercalate "\n" $ ("    " <>) . fmtExpected <$> expecteds,
+          "  but got:",
+          ("    " <>) . fmtExpected $ actual
+        ]
 
 mockNameLabel :: Maybe MockName -> String
 mockNameLabel = maybe mempty (" " <>) . enclose "`"
@@ -506,7 +393,7 @@ data VerifyMatchType a = MatchAny a | MatchAll a
 -- | Class for verifying mock function.
 class Verify params input where
   -- | Verifies that the function has been applied to the expected arguments.
-  shouldApplyTo :: HasCallStack => Mock fun params -> input -> IO ()
+  shouldApplyTo :: HasCallStack => Mock fn params -> input -> IO ()
 
 instance (Eq a, Show a) => Verify (Param a) a where
   shouldApplyTo v a = verify v (MatchAny (param a))
@@ -514,7 +401,7 @@ instance (Eq a, Show a) => Verify (Param a) a where
 instance (Eq a, Show a) => Verify a a where
   shouldApplyTo v a = verify v (MatchAny a)
 
-verify :: (Eq params, Show params) => Mock fun params -> VerifyMatchType params -> IO ()
+verify :: (Eq params, Show params) => Mock fn params -> VerifyMatchType params -> IO ()
 verify (Mock name _ (Verifier ref)) matchType = do
   appliedParamsList <- readAppliedParamsList ref
   let result = doVerify name appliedParamsList matchType
@@ -538,15 +425,36 @@ verifyFailedMessage name appliedParams expected =
     intercalate
       "\n"
       [ "function" <> mockNameLabel name <> " was not applied to the expected arguments.",
-        "  expected: " <> show expected,
+        "  expected: " <> showForMessage (show expected),
         "   but got: " <> formatAppliedParamsList appliedParams
       ]
 
+-- utilities for message formatting
+trim :: String -> String
+trim = f . f
+  where
+    f = reverse . dropWhile (== ' ')
+
+splitByComma :: String -> [String]
+splitByComma s = case break (== ',') s of
+  (a, ',' : rest) -> a : splitByComma rest
+  (a, _) -> [a]
+
 formatAppliedParamsList :: Show a => AppliedParamsList a -> String
 formatAppliedParamsList appliedParams
-  | length appliedParams == 0 = "It has never been applied"
-  | length appliedParams == 1 = init . drop 1 . show $ appliedParams
-  | otherwise = show appliedParams
+  | null appliedParams = "It has never been applied"
+  | length appliedParams == 1 =
+    -- show single element without surrounding list brackets, but quote tokens appropriately
+    let s = show (head appliedParams)
+        inner = if not (null s) && head s == '(' && last s == ')' then init (tail s) else s
+        tokens = map (trim . quoteToken . trim) (splitByComma inner)
+     in intercalate "," tokens
+  | otherwise =
+    -- for multiple applied params, show as a list but ensure tokens are quoted where appropriate
+    let ss = map show appliedParams
+        processed = map (\t -> let inner = if not (null t) && head t == '(' && last t == ')' then init (tail t) else t
+                                in intercalate "," $ map (trim . quoteToken . trim) (splitByComma inner)) ss
+     in show processed
 
 _replace :: Show a => String -> a -> String
 _replace r s = unpack $ replace (pack r) (pack "") (pack (show s))
@@ -565,7 +473,7 @@ class VerifyCount countType params a where
   --   m \`shouldApplyTimes\` (2 :: Int) \`to\` "value" 
   -- @
   --
-  shouldApplyTimes :: HasCallStack => Eq params => Mock fun params -> countType -> a -> IO ()
+  shouldApplyTimes :: HasCallStack => Eq params => Mock fn params -> countType -> a -> IO ()
 
 instance VerifyCount CountVerifyMethod (Param a) a where
   shouldApplyTimes v count a = verifyCount v (param a) count
@@ -600,7 +508,7 @@ compareCount (LessThan e) a = a < e
 compareCount (GreaterThanEqual e) a = a >= e
 compareCount (GreaterThan e) a = a > e
 
-verifyCount :: Eq params => Mock fun params -> params -> CountVerifyMethod -> IO ()
+verifyCount :: Eq params => Mock fn params -> params -> CountVerifyMethod -> IO ()
 verifyCount (Mock name _ (Verifier ref)) v method = do
   appliedParamsList <- readAppliedParamsList ref
   let appliedCount = length (filter (v ==) appliedParamsList)
@@ -632,14 +540,14 @@ class VerifyOrder params input where
   --   print $ stubFn m "b" True
   --   m \`shouldApplyInOrder\` ["a" |\> True, "b" |\> True]
   -- @
-  shouldApplyInOrder :: HasCallStack => Mock fun params -> [input] -> IO ()
+  shouldApplyInOrder :: HasCallStack => Mock fn params -> [input] -> IO ()
 
   -- | Verify that functions are applied in the expected order.
   --
   -- Unlike @'shouldApplyInOrder'@, not all applications need to match exactly.
   --
   -- As long as the order matches, the verification succeeds.
-  shouldApplyInPartialOrder :: HasCallStack => Mock fun params -> [input] -> IO ()
+  shouldApplyInPartialOrder :: HasCallStack => Mock fn params -> [input] -> IO ()
 
 instance (Eq a, Show a) => VerifyOrder (Param a) a where
   shouldApplyInOrder v a = verifyOrder ExactlySequence v $ param <$> a
@@ -654,10 +562,9 @@ data VerifyOrderMethod
   | PartiallySequence
 
 verifyOrder ::
-  Eq params =>
-  Show params =>
+  (Eq params, Show params) =>
   VerifyOrderMethod ->
-  Mock fun params ->
+  Mock fn params ->
   [params] ->
   IO ()
 verifyOrder method (Mock name _ (Verifier ref)) matchers = do
@@ -666,8 +573,7 @@ verifyOrder method (Mock name _ (Verifier ref)) matchers = do
   result & maybe (pure ()) (\(VerifyFailed msg) -> errorWithoutStackTrace msg)
 
 doVerifyOrder ::
-  Eq a =>
-  Show a =>
+  (Eq a, Show a) =>
   VerifyOrderMethod ->
   Maybe MockName ->
   AppliedParamsList a ->
@@ -766,9 +672,8 @@ mapWithIndex f xs = [f i x | (i, x) <- zip [0 ..] xs]
 
 -- | Verify that the function has been applied to the expected arguments at least the expected number of times.
 shouldApplyTimesGreaterThanEqual ::
-  VerifyCount CountVerifyMethod params a =>
-  Eq params =>
-  Mock fun params ->
+  (VerifyCount CountVerifyMethod params a, Eq params) =>
+  Mock fn params ->
   Int ->
   a ->
   IO ()
@@ -776,9 +681,8 @@ shouldApplyTimesGreaterThanEqual m i = shouldApplyTimes m (GreaterThanEqual i)
 
 -- | Verify that the function is applied to the expected arguments less than or equal to the expected number of times.
 shouldApplyTimesLessThanEqual ::
-  VerifyCount CountVerifyMethod params a =>
-  Eq params =>
-  Mock fun params ->
+  (VerifyCount CountVerifyMethod params a, Eq params) =>
+  Mock fn params ->
   Int ->
   a ->
   IO ()
@@ -786,9 +690,8 @@ shouldApplyTimesLessThanEqual m i = shouldApplyTimes m (LessThanEqual i)
 
 -- | Verify that the function has been applied to the expected arguments a greater number of times than expected.
 shouldApplyTimesGreaterThan ::
-  VerifyCount CountVerifyMethod params a =>
-  Eq params =>
-  Mock fun params ->
+  (VerifyCount CountVerifyMethod params a, Eq params) =>
+  Mock fn params ->
   Int ->
   a ->
   IO ()
@@ -796,9 +699,8 @@ shouldApplyTimesGreaterThan m i = shouldApplyTimes m (GreaterThan i)
 
 -- | Verify that the function has been applied to the expected arguments less than the expected number of times.
 shouldApplyTimesLessThan ::
-  VerifyCount CountVerifyMethod params a =>
-  Eq params =>
-  Mock fun params ->
+  (VerifyCount CountVerifyMethod params a, Eq params) =>
+  Mock fn params ->
   Int ->
   a ->
   IO ()
@@ -858,13 +760,13 @@ safeIndex xs n
   | otherwise = listToMaybe (drop n xs)
 
 -- | Verify that it was apply to anything.
-shouldApplyToAnything :: HasCallStack => Mock fun params -> IO ()
+shouldApplyToAnything :: HasCallStack => Mock fn params -> IO ()
 shouldApplyToAnything (Mock name _ (Verifier ref)) = do
   appliedParamsList <- readAppliedParamsList ref
   when (null appliedParamsList) $ error $ "It has never been applied function" <> mockNameLabel name
 
 -- | Verify that it was apply to anything (times).
-shouldApplyTimesToAnything :: Mock fun params -> Int -> IO ()
+shouldApplyTimesToAnything :: Mock fn params -> Int -> IO ()
 shouldApplyTimesToAnything (Mock name _ (Verifier ref)) count = do
   appliedParamsList <- readAppliedParamsList ref
   let appliedCount = length appliedParamsList
@@ -936,4 +838,8 @@ cases a = Cases $ put a
 @casesIO ["a", ""]@ has the same meaning as @cases [ pure \@IO "a", pure \@IO ""]@.
 -}
 casesIO :: [a] -> Cases (IO a) ()
-casesIO = Cases . (put . map pure)
\ No newline at end of file
+casesIO = Cases . (put . map pure)
+
+{-# NOINLINE perform #-}
+perform :: IO a -> a
+perform = unsafePerformIO
\ No newline at end of file
